package com.kakao.auth;

import android.Manifest;
import android.app.Activity;

import com.kakao.auth.authorization.AccessTokenTest;
import com.kakao.auth.authorization.AuthorizationResult;
import com.kakao.auth.authorization.accesstoken.AccessToken;
import com.kakao.test.common.KakaoPowerMockTestCase;
import com.kakao.test.common.TestUtility;
import com.kakao.util.exception.KakaoException;
import com.kakao.util.helper.CommonProtocol;
import com.kakao.util.helper.SharedPreferencesCache;
import com.kakao.util.helper.SystemInfo;
import com.kakao.util.helper.Utility;

import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;
import org.mockito.invocation.InvocationOnMock;
import org.mockito.stubbing.Answer;
import org.powermock.api.mockito.PowerMockito;
import org.powermock.core.classloader.annotations.PrepareForTest;
import org.powermock.reflect.Whitebox;
import org.robolectric.Robolectric;
import org.robolectric.RuntimeEnvironment;
import org.robolectric.util.Transcript;


import java.util.List;


import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;

import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static org.powermock.api.support.membermodification.MemberMatcher.method;
import static org.powermock.api.support.membermodification.MemberModifier.stub;

/**
 * Created by kevin.kang on 2016. 10. 10..
 */

@PrepareForTest({ Session.class, Utility.class, AccessToken.class, KakaoSDK.class })
public class SessionTest extends KakaoPowerMockTestCase {

    private SharedPreferencesCache cache;
    private Session spy;
    private ISessionCallback cb;
    private ISessionCallback cb2;
    private Transcript transcript;

    public enum SessionResults {
        AUTH_CODE_FAILURE,
        AUTH_CODE_SUCCESSS,
        ACCESS_TOKEN_FAILURE,
        ACCESS_TOKEN_SUCCESS,
        OPEN_FAILURE,
        OPEN_SUCCESS
    }
    @Before
    public void setup() {
        super.setUp();

        PowerMockito.spy(Utility.class);
        PowerMockito.spy(AccessToken.class);
        PowerMockito.spy(KakaoSDK.class);
        PowerMockito.spy(Session.class);

        cache = Mockito.spy(new SharedPreferencesCache(RuntimeEnvironment.application.getApplicationContext(), TestUtility.getFakeAppKey()));

        try {
            PowerMockito.whenNew(SharedPreferencesCache.class).withAnyArguments().thenReturn(cache);

            PowerMockito.doReturn(TestUtility.getFakeKeyHash()).when(Utility.class, "getKeyHash", RuntimeEnvironment.application.getApplicationContext());
            PowerMockito.doReturn(TestUtility.getFakeAppKey()).when(Utility.class, "getMetadata", RuntimeEnvironment.application.getApplicationContext(), CommonProtocol.APP_KEY_PROPERTY);

            AccessTokenTest.stubKakaoSDKInit(false);

            stub(method(Session.class, "removeCookiesForKakaoDomain")).toReturn(null);

            PowerMockito.doReturn(true).when(Utility.class, "isUsablePermission", RuntimeEnvironment.application.getApplicationContext(), Manifest.permission.INTERNET);

            Session.initialize(RuntimeEnvironment.application.getApplicationContext(), ApprovalType.INDIVIDUAL, AuthType.KAKAO_LOGIN_ALL);
            spy = PowerMockito.spy(Session.getCurrentSession());
            PowerMockito.doReturn(spy).when(Session.class, "getCurrentSession");

            transcript = new Transcript();
        } catch (Exception e) {
            e.printStackTrace();
            fail();
        }
    }

    @Test
    public void testInitialize() {
        assertNotNull(Session.getCurrentSession());
        assertNotNull(Session.getAppCache());
        assertNotNull(SystemInfo.getKAHeader());
    }

    /**
     * Here denotes the start of testing the actual login mechanism...
     */

    @Test
    public void testOpenWithAuthCodeOAuthErrorResult() {
        final AuthorizationResult result = AuthorizationResult.createAuthCodeOAuthErrorResult("Authorization code not available.");
        final Activity activity = Robolectric.buildActivity(Activity.class).create().get();

        cb = createSessionCallbackWithErrorType(KakaoException.ErrorType.AUTHORIZATION_FAILED);

        spy.addCallback(cb);

        try {
            spyRequestAuthCode(activity, result);
            spy.open(AuthType.KAKAO_TALK, activity);

            assertTrue(spy.isClosed());
            assertFalse(spy.hasValidAccessToken());
        } catch (Exception e) {
            e.printStackTrace();
            fail(e.toString());
        }
    }

    @Test
    public void testOpenWithAuthCodeCancelResult() {
        final AuthorizationResult authCodeResult = AuthorizationResult.createAuthCodeCancelResult("AuthCodeCancelResult");
        final Activity activity = Robolectric.buildActivity(Activity.class).create().get();

        cb = createSessionCallbackWithErrorType(KakaoException.ErrorType.CANCELED_OPERATION);

        try {
            spyRequestAuthCode(activity, authCodeResult);

            spy.open(AuthType.KAKAO_TALK, activity);

            assertTrue(spy.isClosed());
            assertFalse(spy.hasValidAccessToken());
        } catch (Exception e) {
            fail(e.toString());
        }
    }

    @Test
    public void testOpenWithAuthCodePassResult() {
        final AuthorizationResult authCodeResult = AuthorizationResult.createAuthCodePassResult();
        final Activity activity = Robolectric.buildActivity(Activity.class).create().get();

        cb = createSessionCallbackWithErrorType(KakaoException.ErrorType.AUTHORIZATION_FAILED);

        spy.addCallback(cb);

        try {
            spyRequestAuthCode(activity, authCodeResult);
            spyRequestAccessToken(authCodeResult);

            spy.open(AuthType.KAKAO_TALK, activity);

            assertTrue(spy.isClosed());
            assertFalse(spy.hasValidAccessToken());

        } catch (Exception e) {
            fail(e.toString());
        }
    }


    @Test
    public void testOpenWithAccessTokenErrorResult() throws Exception{
        final AuthorizationResult authCodeResult = createSuccessAuthCodeResult();
        final AuthorizationResult accessTokenResult = AuthorizationResult.createAccessTokenErrorResult("AccessToken is not available.");
        final Activity activity = Robolectric.buildActivity(Activity.class).create().get();

        cb = createSessionCallbackWithErrorType(KakaoException.ErrorType.UNSPECIFIED_ERROR);
        spy.addCallback(cb);

        spyRequestAuthCode(activity, authCodeResult);
        spyRequestAccessToken(accessTokenResult);

        spy.open(AuthType.KAKAO_TALK, activity);

        assertTrue(spy.isClosed());
        assertFalse(spy.hasValidAccessToken());
    }

    @Test
    public void testOpenWithAccessTokenOAuthErrorResult() {
        final AuthorizationResult authCodeResult = createSuccessAuthCodeResult();
        final AuthorizationResult accessTokenResult = AuthorizationResult.createAccessTokenOAuthErrorResult("AccessTokenOAuthErrorResult");
        final Activity activity = Robolectric.buildActivity(Activity.class).create().get();

        cb = createSessionCallbackWithErrorType(KakaoException.ErrorType.AUTHORIZATION_FAILED);
        spy.addCallback(cb);

        try {
            spyRequestAuthCode(activity, authCodeResult);
            spyRequestAccessToken(accessTokenResult);

            spy.open(AuthType.KAKAO_TALK, activity);

            assertTrue(spy.isClosed());
            assertFalse(spy.hasValidAccessToken());
        } catch (Exception e) {
            fail(e.toString());
        }
    }


    @Test
    public void testOpenNormal() {
        final AuthorizationResult authCodeResult = createSuccessAuthCodeResult();
        final AuthorizationResult accessTokenResult = AuthorizationResult.createSuccessAccessTokenResult(AccessTokenTest.getStubbedAccessToken());
        final Activity activity = Robolectric.buildActivity(Activity.class).create().get();
        try {
            spyRequestAuthCode(activity, authCodeResult);
            spyRequestAccessToken(accessTokenResult);

            spy.open(AuthType.KAKAO_TALK, activity);

            PowerMockito.verifyPrivate(spy).invoke("requestAuthCode", AuthType.KAKAO_TALK, activity);
            assertTrue(spy.isOpened());
            assertTrue(spy.hasValidAccessToken());
        } catch (Exception e) {
            fail(e.toString());
        }
    }

    @Test
    public void testOpenWhenOpen() {
        setupSessionCallback();
        final AuthorizationResult authCodeResult = createSuccessAuthCodeResult();
        final AuthorizationResult accessTokenResult = AuthorizationResult.createSuccessAccessTokenResult(AccessTokenTest.getStubbedAccessToken());
        final Activity activity = Robolectric.buildActivity(Activity.class).create().get();
        try {
            spyRequestAuthCode(activity, authCodeResult);
            spyRequestAccessToken(accessTokenResult);

            spy.addCallback(cb);
            spy.open(AuthType.KAKAO_TALK, activity);
            assertTrue(spy.isOpened());

            transcript = new Transcript();
            transcript.assertNoEventsSoFar();
            spy.open(AuthType.KAKAO_TALK, activity);

            assertEquals(true, spy.isOpened());
            transcript.assertEventsSoFar(SessionResults.OPEN_SUCCESS.toString());

        } catch (Exception e) {
            fail(e.toString());
        }
    }

    @Test
    public void testCheckStateChange() {
        final AuthorizationResult authCodeResult = createSuccessAuthCodeResult();
        final AuthorizationResult accessTokenResult = AuthorizationResult.createSuccessAccessTokenResult(AccessTokenTest.getStubbedAccessToken());
        final Activity activity = Robolectric.buildActivity(Activity.class).create().get();
        try {
            spyRequestAuthCode(activity, authCodeResult);
            spyRequestAccessToken(accessTokenResult);

            spy.open(AuthType.KAKAO_TALK, activity);

            assertTrue(spy.isOpened());
            assertTrue(spy.hasValidAccessToken());

            AccessToken token = PowerMockito.spy(AccessToken.createEmptyToken());
            PowerMockito.doReturn(false).when(token).hasValidAccessToken();
            PowerMockito.doReturn(true).when(token).hasRefreshToken();
            Whitebox.setInternalState(spy, "accessToken", token);

            spy.checkState();

            assertEquals(true, spy.isOpenable());
            assertEquals(null, Whitebox.getInternalState(spy, "requestType"));

        } catch (Exception e) {
            fail(e.toString());
        }
    }

    @Test
    public void testCheckAccessTokenInfo() {

    }

    @Test
    public void testAddCallback() {
        setupSessionCallback();

        spy.addCallback(cb);
        assertEquals(1, ((List<ISessionCallback>)Whitebox.getInternalState(spy, "sessionCallbacks")).size());

        spy.addCallback(cb);
        assertEquals(1, ((List<ISessionCallback>)Whitebox.getInternalState(spy, "sessionCallbacks")).size());

        spy.addCallback(cb2);
        assertEquals(2, ((List<ISessionCallback>)Whitebox.getInternalState(spy, "sessionCallbacks")).size());

    }

    @Test
    public void testRemoveCallback() {
        ISessionCallback cb = new ISessionCallback() {
            @Override
            public void onSessionOpened() {

            }

            @Override
            public void onSessionOpenFailed(KakaoException exception) {

            }
        };

        ISessionCallback cb2 = new ISessionCallback() {
            @Override
            public void onSessionOpened() {

            }

            @Override
            public void onSessionOpenFailed(KakaoException exception) {

            }
        };

        spy.removeCallback(cb);
        spy.addCallback(cb);
        spy.removeCallback(cb);
        assertEquals(0, ((List<ISessionCallback>)Whitebox.getInternalState(spy, "sessionCallbacks")).size());

        spy.addCallback(cb);
        spy.addCallback(cb2);
        spy.removeCallback(cb2);
        assertEquals(1, ((List<ISessionCallback>)Whitebox.getInternalState(spy, "sessionCallbacks")).size());

    }

    private ISessionCallback createSessionCallbackWithErrorType(final KakaoException.ErrorType errorType) {
        return new ISessionCallback() {
            @Override
            public void onSessionOpened() {
                fail();
            }

            @Override
            public void onSessionOpenFailed(KakaoException exception) {
                assertEquals(errorType, exception.getErrorType());
            }
        };
    }

    private void setupSessionCallback() {
        cb = new ISessionCallback() {
            @Override
            public void onSessionOpened() {
                transcript.add(SessionResults.OPEN_SUCCESS.toString());
            }

            @Override
            public void onSessionOpenFailed(KakaoException exception) {
                transcript.add(SessionResults.OPEN_FAILURE.toString());
            }
        };

        cb2 = new ISessionCallback() {
            @Override
            public void onSessionOpened() {
            }

            @Override
            public void onSessionOpenFailed(KakaoException exception) {

            }
        };
    }

    private AuthorizationResult createSuccessAuthCodeResult() {
        return AuthorizationResult.createSuccessAuthCodeResult(spy.generateRedirectUrl() + "?code=" + TestUtility.getFakeAuthCode());
    }

    private void spyRequestAuthCode(final Activity activity, final AuthorizationResult result) throws Exception {
        PowerMockito.doAnswer(new Answer<Object>() {
            @Override
            public Object answer(InvocationOnMock invocation) throws Throwable {
                spy.onAuthorizationCompletion(result);
                return null;
            }
        }).when(spy, "requestAuthCode", AuthType.KAKAO_TALK, activity);
    }

    private void spyRequestAccessToken(final AuthorizationResult result) throws Exception {
        PowerMockito.doAnswer(new Answer<Object>() {
            @Override
            public Object answer(InvocationOnMock invocation) throws Throwable {
                Whitebox.setInternalState(spy, "requestType", Session.RequestType.GETTING_ACCESS_TOKEN);
                spy.onAccessTokenCompleted(result);
                return null;
            }
        }).when(spy, "requestAccessToken", Session.RequestType.GETTING_ACCESS_TOKEN);
    }
}
