package com.kakao.auth.authorization;

import android.app.Activity;
import android.app.Application;
import android.content.Context;

import com.kakao.auth.ApprovalType;
import com.kakao.auth.AuthType;
import com.kakao.auth.IApplicationConfig;
import com.kakao.auth.ISessionConfig;
import com.kakao.auth.KakaoAdapter;
import com.kakao.auth.KakaoSDK;
import com.kakao.auth.StringSet;
import com.kakao.auth.authorization.accesstoken.AccessToken;
import com.kakao.auth.network.response.AuthResponse;
import com.kakao.network.response.ResponseBody;
import com.kakao.test.common.KakaoPowerMockTestCase;
import com.kakao.util.helper.SharedPreferencesCache;
import com.kakao.util.helper.Utility;

import org.hamcrest.Matchers;
import org.json.JSONObject;
import org.junit.Before;
import org.junit.Test;
import org.powermock.api.mockito.PowerMockito;
import org.powermock.core.classloader.annotations.PrepareForTest;
import org.powermock.reflect.Whitebox;
import org.robolectric.Robolectric;
import org.robolectric.RuntimeEnvironment;


import java.io.IOException;
import java.security.GeneralSecurityException;
import java.util.Date;


import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;

import static org.junit.Assert.assertThat;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static org.powermock.api.support.membermodification.MemberMatcher.method;
import static org.powermock.api.support.membermodification.MemberModifier.stub;

/**
 * This class mocks KakaoSDK, Utility, SharedPreferenceCache, ResponseBody for solely unit-testing AccessToken class.
 * @author kevin.kang
 * Created by kevin.kang on 16. 8. 11..
 */
@PrepareForTest({KakaoSDK.class, Utility.class, AccessToken.class})
public class AccessTokenTest extends KakaoPowerMockTestCase {

    private static final String appKey = "appkey";
    private static final String accessToken = "accessToken";
    private static final String refreshToken = "refreshToken";
    private static final Long expiresIn = 3600 * 12L;
    private static final Long expiresAt = new Date().getTime() + expiresIn;
    private static final String FAKE_KEY_HASH = "lMXltzn4zSwq0EhwLKAo+k0zhqI=";

    private static SharedPreferencesCache cache;
    @Before
    public void setup() {
        super.setUp();

        PowerMockito.spy(AccessToken.class);
        PowerMockito.spy(Utility.class);
        PowerMockito.spy(KakaoSDK.class);

        stubKakaoSDKInit(false);
        cache = PowerMockito.spy(new SharedPreferencesCache(RuntimeEnvironment.application.getApplicationContext(), appKey));

        try {
            PowerMockito.doReturn(accessToken).when(AccessToken.class, "encrypt", RuntimeEnvironment.application.getApplicationContext(), accessToken);
            PowerMockito.doReturn(refreshToken).when(AccessToken.class, "encrypt", RuntimeEnvironment.application.getApplicationContext(), refreshToken);
            PowerMockito.doReturn(accessToken).when(AccessToken.class, "decrypt", RuntimeEnvironment.application.getApplicationContext(), accessToken);
            PowerMockito.doReturn(refreshToken).when(AccessToken.class, "decrypt", RuntimeEnvironment.application.getApplicationContext(), refreshToken);

            PowerMockito.doReturn(FAKE_KEY_HASH).when(Utility.class, "getKeyHash", RuntimeEnvironment.application.getApplicationContext());
        } catch (Exception e) {
            e.printStackTrace();
            fail();
        }
    }

    @Test
    public void testCreateEmptyToken() {
        AccessToken token = AccessToken.createEmptyToken();
        assertNotNull(token);
    }

    @Test
    public void testCreateFromCache() {
        AccessToken token = AccessToken.createFromCache(RuntimeEnvironment.application.getApplicationContext(), cache);
        assertNotNull(token);
        assertFalse(token.hasValidAccessToken());
        assertFalse(token.hasRefreshToken());
    }

    @Test
    public void testSaveAccessTokenToCache() {
        AccessToken token = getStubbedAccessToken();
        if (token == null)
            fail();
        assertEquals(token.getAccessToken(), accessToken);
        assertEquals(token.getRefreshToken(), refreshToken);

        assertThat(token.getRemainedExpiresInAccessTokenTime(), Matchers.lessThanOrEqualTo(3600 * 12 * 1000));

        token.saveAccessTokenToCache(RuntimeEnvironment.application.getApplicationContext(), cache);


        AccessToken tokenFromCache = AccessToken.createFromCache(RuntimeEnvironment.application.getApplicationContext(), cache);
        assertEquals(accessToken, tokenFromCache.getAccessToken());
        assertEquals(refreshToken, tokenFromCache.getRefreshToken());
        assertThat(tokenFromCache.getRemainedExpiresInAccessTokenTime(), Matchers.lessThanOrEqualTo(3600 * 12 * 1000));
    }

    /**
     * This test tests if secure mode is working correctly, but actually bypasses the encryption and decryption mechanism
     * by just using the original value.
     */
    @Test
    public void testSecureMode() {
        stubKakaoSDKInit(true);
        AccessToken token = getStubbedAccessToken();
        if (token == null)
            fail("Failed to get");
        assertEquals(token.getAccessToken(), accessToken);
        assertEquals(token.getRefreshToken(), refreshToken);
        assertThat(token.getRemainedExpiresInAccessTokenTime(), Matchers.lessThanOrEqualTo(3600 * 12 * 1000));

        token.saveAccessTokenToCache(RuntimeEnvironment.application.getApplicationContext(), cache);

        AccessToken tokenFromCache = AccessToken.createFromCache(RuntimeEnvironment.application.getApplicationContext(), cache);

        assertEquals(accessToken, tokenFromCache.getAccessToken());
        assertEquals(refreshToken, tokenFromCache.getRefreshToken());
        assertThat(tokenFromCache.getRemainedExpiresInAccessTokenTime(), Matchers.lessThanOrEqualTo(3600 * 12 * 1000));
    }


    @Test
    public void testTurnOnSecureMode() {
        mockSharedPreferenceCache();
        stubKakaoSDKInit(true);
        stub(method(AccessToken.class, "getLastSecureMode")).toReturn(false);

        AccessToken token = AccessToken.createFromCache(RuntimeEnvironment.application.getApplicationContext(), cache);
        assertTrue(token.hasValidAccessToken());
        assertTrue(token.hasRefreshToken());
    }

    @Test
    public void testTurnOffSecureMode() {
        mockSharedPreferenceCache();
        stubKakaoSDKInit(false);
        stub(method(AccessToken.class, "getLastSecureMode")).toReturn(true);

        AccessToken token = AccessToken.createFromCache(RuntimeEnvironment.application.getApplicationContext(), cache);
        assertTrue(token.hasValidAccessToken());
        assertTrue(token.hasRefreshToken());
    }

    @Test
    public void testGeneralSecurityException() {
        mockSharedPreferenceCache();
        stubKakaoSDKInit(true);

        try {
            PowerMockito.doThrow(new GeneralSecurityException()).when(AccessToken.class, "encrypt", org.mockito.Matchers.any(Application.class), org.mockito.Matchers.anyString());
            PowerMockito.doThrow(new GeneralSecurityException()).when(AccessToken.class, "decrypt", org.mockito.Matchers.any(Application.class), org.mockito.Matchers.anyString());
        } catch (Exception e) {
            e.printStackTrace();
            fail(e.toString());
        }

        AccessToken token = AccessToken.createFromCache(RuntimeEnvironment.application.getApplicationContext(), cache);

        assertFalse(token.hasValidAccessToken());
        assertFalse(token.hasRefreshToken());
    }

    @Test
    public void testIOException() {
        mockSharedPreferenceCache();
        stubKakaoSDKInit(true);

        try {
            PowerMockito.doThrow(new IOException()).when(AccessToken.class, "encrypt", org.mockito.Matchers.any(Application.class), org.mockito.Matchers.anyString());
            PowerMockito.doThrow(new IOException()).when(AccessToken.class, "decrypt", org.mockito.Matchers.any(Application.class), org.mockito.Matchers.anyString());
        } catch (Exception e) {
            e.printStackTrace();
            fail(e.toString());
        }

        AccessToken token = AccessToken.createFromCache(RuntimeEnvironment.application.getApplicationContext(), cache);

        assertFalse(token.hasValidAccessToken());
        assertFalse(token.hasRefreshToken());
    }

    /**
     * Below are private helper methods
     */

    /**
     * Mocks an access token built from auth API response.
     * @return
     */
    public static AccessToken getStubbedAccessToken() {
        ResponseBody spy;
        try {
            spy = PowerMockito.spy(new ResponseBody(200, new JSONObject()));

            PowerMockito.doReturn(true).when(spy).has(StringSet.access_token);
            PowerMockito.doReturn(true).when(spy).has(StringSet.refresh_token);
            PowerMockito.doReturn(accessToken).when(spy).getString(StringSet.access_token);
            PowerMockito.doReturn(refreshToken).when(spy).getString(StringSet.refresh_token);
            PowerMockito.doReturn(3600 * 12).when(spy).getInt(StringSet.expires_in);
            return new AccessToken(spy);
        } catch (ResponseBody.ResponseBodyException|AuthResponse.AuthResponseStatusError e) {
            e.printStackTrace();
            return null;
        }
    }

    /**
     * Mocks SharedPreference cache to build an access token on app start.
     */
    public static void mockSharedPreferenceCache() {
        if (cache == null)
            return;

        PowerMockito.doReturn(accessToken).when(cache).getString(Whitebox.getInternalState(AccessToken.class, "CACHE_ACCESS_TOKEN").toString());
        PowerMockito.doReturn(refreshToken).when(cache).getString(Whitebox.getInternalState(AccessToken.class, "CACHE_REFRESH_TOKEN").toString());
        PowerMockito.doReturn(expiresAt).when(cache).getLong(Whitebox.getInternalState(AccessToken.class, "CACHE_ACCESS_TOKEN_EXPIRES_AT").toString());
        PowerMockito.doReturn(expiresAt).when(cache).getLong(Whitebox.getInternalState(AccessToken.class, "CACHE_REFRESH_TOKEN_EXPIRES_AT").toString());
    }

    public static void stubKakaoSDKInit(final boolean isSecureMode) {
        final Activity activity = Robolectric.buildActivity(Activity.class).create().get();
        KakaoAdapter adapter = new KakaoAdapter() {
            @Override
            public IApplicationConfig getApplicationConfig() {
                return new IApplicationConfig() {
                    @Override
                    public Context getApplicationContext() {
                        return RuntimeEnvironment.application.getApplicationContext();
                    }
                };
            }

            @Override
            public ISessionConfig getSessionConfig() {
                return new ISessionConfig() {
                    @Override
                    public AuthType[] getAuthTypes() {
                        return new AuthType[0];
                    }

                    @Override
                    public boolean isUsingWebviewTimer() {
                        return false;
                    }

                    @Override
                    public boolean isSecureMode() {
                        return isSecureMode;
                    }

                    @Override
                    public ApprovalType getApprovalType() {
                        return null;
                    }

                    @Override
                    public boolean isSaveFormData() {
                        return false;
                    }
                };
            }
        };

        try {
            PowerMockito.doReturn(adapter).when(KakaoSDK.class, "getAdapter");
        } catch (Exception e) {
            e.printStackTrace();
            fail();
        }
    }
}
