package com.kakao.auth.authorization.authcode;

import android.app.Activity;
import android.content.Intent;

import com.kakao.auth.AuthType;
import com.kakao.auth.StartActivityWrapper;
import com.kakao.auth.authorization.AuthorizationResult;
import com.kakao.auth.authorization.Authorizer;
import com.kakao.test.common.KakaoPowerMockTestCase;
import com.kakao.test.common.TestUtility;
import com.kakao.util.exception.KakaoException;
import com.kakao.util.helper.KakaoServiceProtocol;
import com.kakao.util.helper.TalkProtocol;

import org.junit.Before;
import org.junit.Test;
import org.powermock.api.mockito.PowerMockito;
import org.powermock.core.classloader.annotations.PrepareForTest;
import org.powermock.reflect.Whitebox;
import org.robolectric.Robolectric;
import org.robolectric.RuntimeEnvironment;

import java.util.Queue;

import static org.junit.Assert.assertEquals;

import static org.junit.Assert.fail;
import static org.powermock.api.support.membermodification.MemberMatcher.method;

/**
 * Created by kevin.kang on 2016. 10. 10..
 */
@PrepareForTest({ GetterAuthCode.class, AuthCodeRequest.class, KakaoServiceProtocol.class })
public class GetterAuthCodeTest extends KakaoPowerMockTestCase {
    private GetterAuthCode getterAuthCode;
    private Activity activity;
    private AuthCodeRequest request;
    private Queue<AuthCodeRequest.Command> commandQueue;
    @Before
    public void setup() {
        super.setUp();

        PowerMockito.spy(GetterAuthCode.class);
        activity = Robolectric.buildActivity(Activity.class).create().get();
        request = PowerMockito.spy(new AuthCodeRequest(RuntimeEnvironment.application.getApplicationContext(), TestUtility.getFakeAppKey(), "redirect_url"));
    }

    @Test
    public void testKakaoTalk() {
        getterAuthCode = GetterAuthCode.newInstance(request, AuthType.KAKAO_TALK, new StartActivityWrapper(activity));
        commandQueue = Whitebox.getInternalState(getterAuthCode, "commandQueue");
        assertEquals(3, commandQueue.size());

    }

    @Test
    public void testKakaoStory() {
        getterAuthCode = GetterAuthCode.newInstance(request, AuthType.KAKAO_STORY, new StartActivityWrapper(activity));
        commandQueue = Whitebox.getInternalState(getterAuthCode, "commandQueue");
        assertEquals(2, commandQueue.size());

    }

    @Test
    public void testExcludeNativeLogin() {
        getterAuthCode = GetterAuthCode.newInstance(request, AuthType.KAKAO_TALK_EXCLUDE_NATIVE_LOGIN, new StartActivityWrapper(activity));
        commandQueue = Whitebox.getInternalState(getterAuthCode, "commandQueue");
        assertEquals(2, commandQueue.size());

    }

    @Test
    public void testKakaoAcount() {
        getterAuthCode = GetterAuthCode.newInstance(request, AuthType.KAKAO_ACCOUNT, new StartActivityWrapper(activity));
        commandQueue = Whitebox.getInternalState(getterAuthCode, "commandQueue");
        assertEquals(1, commandQueue.size());
    }

    @Test
    public void testKakaoLoginAll() {
        getterAuthCode = GetterAuthCode.newInstance(request, AuthType.KAKAO_LOGIN_ALL, new StartActivityWrapper(activity));
        commandQueue = Whitebox.getInternalState(getterAuthCode, "commandQueue");
        assertEquals(4, commandQueue.size());

    }

    @Test
    public void testDoneOnOAuthError() {
        getterAuthCode = GetterAuthCode.newInstance(request, AuthType.KAKAO_TALK, new StartActivityWrapper(activity));
        commandQueue = Whitebox.getInternalState(getterAuthCode, "commandQueue");

        getterAuthCode.doneOnOAuthError("Error message");

        assertEquals(0, commandQueue.size());
    }

    @Test
    public void testStartWithNullIntent() {
        getterAuthCode = GetterAuthCode.newInstance(request, AuthType.KAKAO_TALK, new StartActivityWrapper(activity));
        commandQueue = Whitebox.getInternalState(getterAuthCode, "commandQueue");

        Whitebox.setInternalState(getterAuthCode, "hasInternetPermission", true);

        getterAuthCode.start();

        try {
            PowerMockito.verifyPrivate(activity).invoke(method(Activity.class, "startActivityForResult", Intent.class, int.class));
        } catch (Exception e) {
            fail(e.toString());
        }
    }

    @Test
    public void testStartWithIntent() {
        getterAuthCode = GetterAuthCode.newInstance(request, AuthType.KAKAO_TALK, new StartActivityWrapper(activity));
        commandQueue = Whitebox.getInternalState(getterAuthCode, "commandQueue");

        Whitebox.setInternalState(getterAuthCode, "hasInternetPermission", true);
        PowerMockito.doReturn(new Intent()).when(request).getIntent(AuthCodeRequest.Command.LOGGED_IN_TALK);

        getterAuthCode.start();

        try {
            PowerMockito.verifyPrivate(activity).invoke(method(Activity.class, "startActivityForResult", Intent.class, int.class));
        } catch (Exception e) {
            fail(e.toString());
        }
    }

//    @Test
//    public void testStartWith

    @Test
    public void testOnWebViewCompletedWithIllegalArgument() {
        getterAuthCode = GetterAuthCode.newInstance(request, AuthType.KAKAO_TALK, new StartActivityWrapper(activity));
        commandQueue = Whitebox.getInternalState(getterAuthCode, "commandQueue");

        getterAuthCode.onWebViewCompleted("redirect url", new KakaoException(KakaoException.ErrorType.ILLEGAL_ARGUMENT, "Error message"));

        assertEquals(0, commandQueue.size());

        Authorizer.OnAuthorizationListener listener = Whitebox.getInternalState(getterAuthCode, "onAuthorizationListener");
        try {
            PowerMockito.verifyPrivate(listener).invoke(method(Authorizer.OnAuthorizationListener.class, "onAuthorizationCompletion", AuthorizationResult.class));
        } catch (Exception e) {
            fail(e.toString());
        }
    }

    @Test
    public void testOnWebViewCompletedWithMisconfiguration() {
        getterAuthCode = GetterAuthCode.newInstance(request, AuthType.KAKAO_TALK, new StartActivityWrapper(activity));
        commandQueue = Whitebox.getInternalState(getterAuthCode, "commandQueue");

        getterAuthCode.onWebViewCompleted(null, new KakaoException(KakaoException.ErrorType.MISS_CONFIGURATION, "Error message"));
    }

    @Test
    public void testOnWebViewCompletedWithCanceledOperation() {
        getterAuthCode = GetterAuthCode.newInstance(request, AuthType.KAKAO_TALK, new StartActivityWrapper(activity));
        commandQueue = Whitebox.getInternalState(getterAuthCode, "commandQueue");

        getterAuthCode.onWebViewCompleted(null, new KakaoException(KakaoException.ErrorType.CANCELED_OPERATION, "Error message"));
    }

    @Test
    public void testOnWebViewCompletedWithFailedAuthorization() {
        getterAuthCode = GetterAuthCode.newInstance(request, AuthType.KAKAO_TALK, new StartActivityWrapper(activity));
        commandQueue = Whitebox.getInternalState(getterAuthCode, "commandQueue");

        getterAuthCode.onWebViewCompleted(null, new KakaoException(KakaoException.ErrorType.AUTHORIZATION_FAILED, "Error message"));
    }

    @Test
    public void testOnWebViewCompletedWithUnspecifiedError() {
        getterAuthCode = GetterAuthCode.newInstance(request, AuthType.KAKAO_TALK, new StartActivityWrapper(activity));
        commandQueue = Whitebox.getInternalState(getterAuthCode, "commandQueue");

        getterAuthCode.onWebViewCompleted(null, new KakaoException(KakaoException.ErrorType.UNSPECIFIED_ERROR, "Error message"));
    }

    @Test
    public void testHandleActivityResult() {
        initGetterAuthCode();

        Intent intent = new Intent();
        intent.putExtra(TalkProtocol.EXTRA_REDIRECT_URL, "redirect_url");
        getterAuthCode.handleActivityResult(0, Activity.RESULT_OK, intent);

        assertEquals(0, commandQueue.size());

        try {
            Authorizer.OnAuthorizationListener listener = Whitebox.getInternalState(getterAuthCode, "onAuthorizationListener");
            PowerMockito.verifyPrivate(listener).invoke(method(Authorizer.OnAuthorizationListener.class, "onAuthorizationCompletion", AuthorizationResult.class));
        } catch (Exception e) {
            fail(e.toString());
        }
    }

    @Test
    public void testHandleActivityResultWithNullIntent() {
        initGetterAuthCode();
        getterAuthCode.handleActivityResult(0, 0, null);

        assertEquals(0, commandQueue.size());

        try {
            Authorizer.OnAuthorizationListener listener = Whitebox.getInternalState(getterAuthCode, "onAuthorizationListener");
            try {
                PowerMockito.verifyPrivate(listener).invoke(method(Authorizer.OnAuthorizationListener.class, "onAuthorizationCompletion", AuthorizationResult.class));
            } catch (Exception e) {
                fail(e.toString());
            }        } catch (Exception e) {
            fail(e.toString());
        }
    }

    @Test
    public void testHandleActivityResultWithDifferentProtocol() {
        initGetterAuthCode();
        PowerMockito.spy(KakaoServiceProtocol.class);

        Intent intent = new Intent();
        try {
            PowerMockito.doReturn(false).when(KakaoServiceProtocol.class, "isCapriProtocolMatched", intent);
            getterAuthCode.handleActivityResult(0, 0, intent);

            assertEquals(0, commandQueue.size());

            Authorizer.OnAuthorizationListener listener = Whitebox.getInternalState(getterAuthCode, "onAuthorizationListener");
            PowerMockito.verifyPrivate(listener).invoke(method(Authorizer.OnAuthorizationListener.class, "onAuthorizationCompletion", AuthorizationResult.class));
        } catch (Exception e) {
            fail(e.toString());
        }
    }

    @Test
    public void testHandleActivityResultWithDifferentResultCode() {
        initGetterAuthCode();

        Intent intent = new Intent();

        try {
            getterAuthCode.handleActivityResult(0, Activity.RESULT_FIRST_USER, intent);

            assertEquals(0, commandQueue.size());
            Authorizer.OnAuthorizationListener listener = Whitebox.getInternalState(getterAuthCode, "onAuthorizationListener");
            PowerMockito.verifyPrivate(listener).invoke(method(Authorizer.OnAuthorizationListener.class, "onAuthorizationCompletion", AuthorizationResult.class));
        } catch (Exception e) {
            fail(e.toString());
        }
    }
    private void initGetterAuthCode() {
        getterAuthCode = GetterAuthCode.newInstance(request, AuthType.KAKAO_TALK, new StartActivityWrapper(activity));
        commandQueue = Whitebox.getInternalState(getterAuthCode, "commandQueue");
    }
}
