/*
  Copyright 2016-2017 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.kakaolink.v2;

import android.app.AlertDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.net.Uri;
import android.os.Handler;
import android.os.Looper;

import com.kakao.kakaolink.R;
import com.kakao.kakaolink.internal.KakaoTalkLinkProtocol;
import com.kakao.kakaolink.v2.network.KakaoLinkTemplateRequest;
import com.kakao.kakaolink.v2.network.LinkImageDeleteRequest;
import com.kakao.kakaolink.v2.network.LinkImageScrapRequest;
import com.kakao.kakaolink.v2.network.LinkImageUploadRequest;
import com.kakao.kakaolink.v2.network.TemplateDefaultRequest;
import com.kakao.kakaolink.v2.network.TemplateScrapRequest;
import com.kakao.kakaolink.v2.network.TemplateScrapResponse;
import com.kakao.kakaolink.v2.network.TemplateValidateRequest;
import com.kakao.kakaolink.v2.network.KakaoLinkTemplateResponse;
import com.kakao.network.ErrorResult;
import com.kakao.network.callback.ResponseCallback;
import com.kakao.network.storage.ImageDeleteResponse;
import com.kakao.network.storage.ImageUploadResponse;
import com.kakao.network.tasks.KakaoResultTask;
import com.kakao.network.tasks.KakaoTaskQueue;
import com.kakao.util.exception.KakaoException;
import com.kakao.util.helper.log.Logger;

import org.json.JSONObject;

/**
 * 실제로 Template validate을 하고 Intent를 생성하여 카카오링크 4.0 메시지를 보내는 클래스.
 * @author kevin.kang
 * Created by kevin.kang on 2016. 11. 28..
 */

class KakaoLinkSender {
    private final KakaoTaskQueue taskQueue;
    private final KakaoLinkApi linkApi;

    KakaoLinkSender(final KakaoTaskQueue taskQueue, final KakaoLinkApi linkApi) {
        this.taskQueue = taskQueue;
        this.linkApi = linkApi;
    }

    void send(final Context context, final KakaoLinkTemplateRequest request, final ResponseCallback<KakaoLinkResponse> callback) {
        taskQueue.addTask(getKakaoLinkResultTask(context, request, callback));
    }

    void uploadImage(final LinkImageUploadRequest request, final ResponseCallback<ImageUploadResponse> callback) {
        taskQueue.addTask(getImageUploadResultTask(request, callback));
    }

    void uploadImageAfterScrap(final LinkImageScrapRequest request, final ResponseCallback<ImageUploadResponse> callback) {
        taskQueue.addTask(getImageScrapResultTask(request, callback));
    }

    void deleteImage(final LinkImageDeleteRequest request, final ResponseCallback<ImageDeleteResponse> callback) {
        taskQueue.addTask(getImageDeleteResultTask(request, callback));
    }

    @SuppressWarnings("WeakerAccess")
    KakaoResultTask<KakaoLinkResponse> getKakaoLinkResultTask(final Context context, final KakaoLinkTemplateRequest request, final ResponseCallback<KakaoLinkResponse> callback) {
        return new KakaoResultTask<KakaoLinkResponse>(callback) {
            @Override
            public KakaoLinkResponse call() throws Exception {
                KakaoLinkTemplateResponse response;
                if (request instanceof TemplateValidateRequest) {
                    response = linkApi.requestTemplateValidate((TemplateValidateRequest)request);
                } else if (request instanceof TemplateDefaultRequest) {
                    response = linkApi.requestTemplateDefault((TemplateDefaultRequest)request);
                } else {
                    response = linkApi.requestTemplateScrap((TemplateScrapRequest)request);
                }

                Intent intent;
                try {
                    intent = createKakaoLinkIntent(request, response);
                    context.startActivity(intent);
                    return new KakaoLinkResponse(response.getTemplateMsg(), response.getWarningMsg(), response.getArgumentMsg());
                } catch (final KakaoException e) {
                    if (e.getErrorType() == KakaoException.ErrorType.URI_LENGTH_EXCEEDED) {
                        final KakaoException eWithMessage = new KakaoException(KakaoException.ErrorType.URI_LENGTH_EXCEEDED, context.getResources().getString(R.string.com_kakao_alert_uri_too_long));
                        Handler handler = new Handler(Looper.getMainLooper());
                        handler.post(new Runnable() {
                            @Override
                            public void run() {
                                getUriLengthExceededAlertDialog(context, eWithMessage, callback).show();
                            }
                        });
                        throw eWithMessage;
                    } else {
                        throw new KakaoException(KakaoException.ErrorType.UNSPECIFIED_ERROR);
                    }
                }
            }
        };
    }

    @SuppressWarnings("WeakerAccess")
    KakaoResultTask<ImageUploadResponse> getImageUploadResultTask(final LinkImageUploadRequest request, final ResponseCallback<ImageUploadResponse> callback) {
        return new KakaoResultTask<ImageUploadResponse>(callback) {
            @Override
            public ImageUploadResponse call() throws Exception {
                return linkApi.requestImageUpload(request);
            }
        };
    }

    KakaoResultTask<ImageUploadResponse> getImageScrapResultTask(final LinkImageScrapRequest request, final ResponseCallback<ImageUploadResponse> callback) {
        return new KakaoResultTask<ImageUploadResponse>(callback) {
            @Override
            public ImageUploadResponse call() throws Exception {
                return linkApi.requestImageScrap(request);
            }
        };
    }

    KakaoResultTask<ImageDeleteResponse> getImageDeleteResultTask(final LinkImageDeleteRequest request, final ResponseCallback<ImageDeleteResponse> callback) {
        return new KakaoResultTask<ImageDeleteResponse>(callback) {
            @Override
            public ImageDeleteResponse call() throws Exception {
                return linkApi.requestImageDelete(request);
            }
        };
    }

    Intent createKakaoLinkIntent(final KakaoLinkTemplateRequest request, final KakaoLinkTemplateResponse response) {
        Uri.Builder builder = new Uri.Builder();
        builder.scheme(KakaoTalkLinkProtocol.LINK_SCHEME).authority(KakaoTalkLinkProtocol.LINK_AUTHORITY);
        builder.appendQueryParameter(KakaoTalkLinkProtocol.LINKVER, KakaoTalkLinkProtocol.LINK_VERSION_40);
        builder.appendQueryParameter(KakaoTalkLinkProtocol.APP_KEY, request.getAppKey());
        builder.appendQueryParameter(KakaoTalkLinkProtocol.APP_VER, request.getAppVer());


        String templateId;
        String templateArgsString = null;

        if (response instanceof TemplateScrapResponse) {
            templateId = ((TemplateScrapResponse) response).getTemplateId();
        } else {
            templateId = request.getTemplateId();
        }
        if (templateId != null) {
            builder.appendQueryParameter(KakaoTalkLinkProtocol.TEMPLATE_ID, templateId);
        }

        if (response instanceof TemplateScrapResponse) {
            JSONObject templateArgs = ((TemplateScrapResponse) response).getTemplateArgs();
            if (templateArgs != null) {
                templateArgsString = templateArgs.toString();
            }
        } else  {
            templateArgsString = request.getTemplateArgsString();
        }
        if (templateArgsString != null) {
            builder.appendQueryParameter(KakaoTalkLinkProtocol.TEMPLATE_ARGS, templateArgsString);
        }

        builder.appendQueryParameter(KakaoTalkLinkProtocol.TEMPLATE_JSON, response.getTemplateMsg().toString());
        Uri uri = builder.build();
        if (uri.toString().length() > KakaoTalkLinkProtocol.LINK_URI_MAX_SIZE) {
            throw new KakaoException(KakaoException.ErrorType.URI_LENGTH_EXCEEDED);
        }
        Intent intent = new Intent(Intent.ACTION_VIEW, uri);
        intent.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
        return intent;
    }

    @SuppressWarnings("WeakerAccess")
    AlertDialog getUriLengthExceededAlertDialog(final Context context, final KakaoException e, final ResponseCallback<KakaoLinkResponse> callback) {
        String message = context.getResources().getString(R.string.com_kakao_alert_uri_too_long);
        AlertDialog dialog = new AlertDialog.Builder(context)
                .setIcon(android.R.drawable.ic_dialog_alert)
                .setMessage(message)
                .setPositiveButton(android.R.string.ok, new DialogInterface.OnClickListener() {
                    @Override
                    public void onClick(DialogInterface dialog, int which) {
                    }
                }).create();

        dialog.setOnDismissListener(new DialogInterface.OnDismissListener() {
            @Override
            public void onDismiss(DialogInterface dialog) {
                callback.onHandledFailure(new ErrorResult(e));
            }
        });
        return dialog;
    }
}